import asedftk

import pytest

from pytest import approx

from ase.io import read, write
from ase.build import bulk, molecule

#
# Adapted from the ASE test suite code in
# https://gitlab.com/ase/ase/-/blob/master/ase/test/calculator/test_al.py
#
@pytest.mark.skip("Collinear spin not yet in DFTK")
def test_aluminium_oxide():
    ENERGY_LDA = 0.0
    ENERGY_PBE = 0.0

    kpts = [2, 2, 2]
    params = {"scftol": 0.0001, "ecut": 200}

    calc = asedftk.DFTK(label="dftk", xc="LDA", kpts=kpts, **params)
    al = bulk("AlO", crystalstructure='rocksalt', a=4.5)
    al.calc = calc
    elda = al.get_potential_energy()
    assert elda == approx(ENERGY_LDA, rel=5e-3)

    calc.set(xc='PBE', kpts=kpts)
    epbe = al.get_potential_energy()
    assert not calc.calculation_required(al, ['energy'])
    assert epbe == approx(ENERGY_PBE, rel=5e-3)

    al = calc.get_atoms()
    elda2 = al.get_potential_energy()
    assert elda2 == approx(ENERGY_LDA, rel=5e-3)

    label = "dir/dftk-2"
    calc = asedftk.DFTK(label=label, atoms=al, xc="LDA", kpts=kpts, **params)
    elda3 = al.get_potential_energy()
    assert elda == elda3
    assert elda3 == approx(ENERGY_LDA, rel=5e-3)


#
# Adapted from the ASE test suite code in
# https://gitlab.com/ase/ase/-/blob/master/ase/test/calculator/test_h2.py
#
def test_hydrogen():
    ENERGY_H1_LDA = 0.0
    ENERGY_H1_PBE = 0.0
    ENERGY_H2_LDA = -29.56773881096904
    ENERGY_H2_PBE = -30.320772272053382
    BINIDNG_LDA = 2 * ENERGY_H1_LDA - ENERGY_H2_LDA
    BINIDNG_PBE = 2 * ENERGY_H1_PBE - ENERGY_H2_PBE

    kpts = [2, 2, 2]
    params = {"scftol": 0.0001, "ecut": 200}

    # H2 molecule
    calc = asedftk.DFTK(label="dftk", xc="LDA", kpts=kpts, **params)
    h2 = molecule('H2', calculator=calc, pbc=True)
    h2.center(vacuum=2.0)
    e2 = h2.get_potential_energy()
    calc.set(xc='PBE')
    e2pbe = h2.get_potential_energy()
    assert e2 == approx(ENERGY_H2_LDA, rel=5e-3)
    assert e2pbe == approx(ENERGY_H2_PBE, rel=5e-3)

    # TODO Remove this once we are there
    pytest.skip("DFTK does not yet have collinear spin.")

    # H atom
    h1 = h2.copy()
    del h1[1]
    h1.set_initial_magnetic_moments([1])
    h1.calc = calc
    e1pbe = h1.get_potential_energy()
    calc.set(xc='LDA')
    e1 = h1.get_potential_energy()
    assert e1 == approx(ENERGY_H1_LDA, rel=5e-3)
    assert e1pbe == approx(ENERGY_H1_PBE, rel=5e-3)

    try:
        m1 = h1.get_magnetic_moment()
    except NotImplementedError:
        pass
    else:
        assert m1 == 0.0

    # Check binding energy
    assert 2 * e1 - e2 == approx(BINIDNG_LDA, rel=5e-3)
    assert 2 * e1pbe - e2pbe == approx(BINIDNG_PBE, rel=5e-3)

    # Check some restart workflow
    calc = asedftk.DFTK(restart="dftk")
    assert calc.atoms == h1
    assert not calc.calculation_required(h1, ['energy'])
    h1 = calc.get_atoms()
    assert h1.get_potential_energy() == e1

    calc = asedftk.DFTK(label="dir/dftk-h1", atoms=h1, xc='LDA', kpts=kpts, **params)
    assert h1.get_potential_energy() == e1


#
# Adapted from the ASE test suite code in
# https://gitlab.com/ase/ase/-/blob/master/ase/test/calculator/test_traj.py
#
def test_trajectories():
    ENERGY = -30.38758246199375
    FORCES = [[0.0, 0.0,  1.88397764],
              [0.0, 0.0, -1.88397764]]

    h2 = molecule('H2', pbc=True)
    h2.center(vacuum=2.0)
    h2.calc = asedftk.DFTK()
    e = h2.get_potential_energy()
    assert not h2.calc.calculation_required(h2, ['energy'])
    assert e == approx(ENERGY, rel=5e-5, abs=1e-6)
    f = h2.get_forces()
    assert not h2.calc.calculation_required(h2, ['energy', 'forces'])
    for i in range(len(FORCES)):
        assert f[i] == approx(FORCES[i], rel=1e-2, abs=1e-3)

    # Writing and reading trajectories
    write('h2.traj', h2)
    h2 = read('h2.traj')
    assert abs(e - h2.get_potential_energy()) < 1e-12
    assert abs(f - h2.get_forces()).max() < 1e-12



__version__ = "0.2.1"
__license__ = "MIT"
__author__ = ["Michael F. Herbst"]

import io
import os
import json
import datetime
import subprocess

import ase.io

from ase.calculators.calculator import (CalculationFailed, Calculator,
                                        CalculatorSetupError, Parameters,
                                        all_changes, register_calculator_class)

__all__ = ["DFTK", "update"]


def environment():
    dftk_environment = os.path.join(os.path.dirname(__file__), "dftk_environment")
    return os.environ.get("ASEDFTK_DFTK_ENVIRONMENT", dftk_environment)


def julia(*args, n_mpi=1, **kwargs):
    mpiargs = []
    if n_mpi > 1:
        raise NotImplementedError("MPI not yet implemented. Please select n_mpi=1.")
        mpiargs = [get_mpiexecjl(), "--project=" + environment(), "-np", str(n_mpi)]

    julia_exe = os.environ.get("JULIA", "julia")
    try:
        args = mpiargs + [julia_exe, "--project=" + environment(),
                          "--startup-file=no", *args]
        return subprocess.check_call(args, **kwargs)
    except FileNotFoundError:
        raise RuntimeError(
            "Julia not found. Please check that Julia is installed and "
            "added to your PATH variable or otherwise point the JULIA "
            "environment variable to the full path of your julia binary."
        )


def check_julia_version(min_version="1.5.0"):
    try:
        return julia("-e", f'VERSION < v"{min_version}" && Sys.exit(1)')
    except subprocess.CalledProcessError:
        raise RuntimeError(
            f"Julia version below minimal version {min_version}. "
            "Please upgrade."
        )


def get_mpiexecjl():
    # XXX More flexible is to use 'joinpath(DEPOT_PATH[1], "bin", "mpiexecjl")',
    # which is used by MPI to determine the install location.
    return os.path.join(os.path.expanduser("~"), ".julia", "bin", "mpiexecjl")


def update(always_run=True):
    """
    Update the installed Julia dependencies.
    """
    updated_file = os.path.join(environment(), "last_updated")
    project_file = os.path.join(environment(), "Project.toml")

    # Update file is older than Project file
    needs_update = (not os.path.isfile(updated_file)
                    or os.stat(updated_file).st_mtime < os.stat(project_file).st_mtime)

    if not needs_update and not always_run:
        return

    check_julia_version()
    julia("-e", "import Pkg; Pkg.instantiate(); Pkg.update(); Pkg.precompile()")
    open(updated_file, "w").close()

    if not os.path.isfile(get_mpiexecjl()):
        julia("-e", "import MPI; MPI.install_mpiexecjl(verbose=true)")
        assert os.path.isfile(get_mpiexecjl())


def run_calculation(properties, inputfile, n_threads=1, n_mpi=1):
    check_julia_version()
    update(always_run=False)
    script = os.path.join(os.path.dirname(__file__), "run_calculation.jl")
    logfile = os.path.splitext(inputfile)[0] + ".log"
    if n_threads is None:
        try:
            n_threads = int(os.environ.get("JULIA_NUM_THREADS", "1"))
        except ValueError:
            n_threads = 1

    try:
        with open(logfile, "a") as fp:
            fp.write(f"#\n#--  {datetime.datetime.now()}\n#\n")
            fp.flush()
            julia(*["-t", str(n_threads), script, *properties, inputfile],
                  n_mpi=n_mpi, stderr=subprocess.STDOUT, stdout=fp)
            fp.flush()
            fp.write("\n")
    except subprocess.CalledProcessError:
        msg = f"DFTK calculation failed. See logfile {logfile} for details."
        if "CI" in os.environ:
            with open(logfile, "r") as fp:
                msg += "\n\n" + fp.read()
        raise CalculationFailed(msg)


class DFTK(Calculator):
    implemented_properties = ["energy", "forces"]

    def __init__(self, atoms=None, label="dftk", **kwargs):
        self.default_parameters = {
            "xc": "LDA",
            "kpts": 3.5,
            "smearing": None,
            "nbands": None,
            "charge": 0.0,
            "functionals": None,
            "pps": "hgh",
            "scftol": 1e-5,
            "ecut": 400,
            "mixing": None,
            "n_mpi": 1,
            "n_threads": None,
        }
        self.scfres = label + ".scfres.jld2"
        super().__init__(label=label, atoms=atoms, **kwargs)

    def reset(self):
        # Reset internal state by purging all intermediates
        super().reset()
        if os.path.isfile(self.scfres):
            os.unlink(self.scfres)

    def read(self, label):
        super().read(label)

        with open(label + ".json", "r") as fp:
            saved_dict = json.load(fp)
        self.parameters = Parameters(saved_dict["parameters"])
        self.results = saved_dict["results"]
        self.scfres = saved_dict.get("scfres", None)

        with io.StringIO(saved_dict["atoms"]) as fp:
            self.atoms = ase.io.read(fp, format="json")

    def write(self, label=None):
        if label is None:
            label = self.label
        if self.atoms is None:
            raise CalculatorSetupError("An Atoms object must be present "
                                       "in the calculator to use write")

        with io.StringIO() as fp:
            ase.io.write(fp, self.atoms.copy(), format="json")
            atoms_json = fp.getvalue()

        with open(label + ".json", "w") as fp:
            save_dict = {
                "parameters": self.parameters,
                "results": self.results,
                "atoms": atoms_json,
                "scfres": self.scfres,
            }
            json.dump(save_dict, fp)

    def set(self, **kwargs):
        changed_parameters = super().set(**kwargs)
        if changed_parameters:
            # Currently reset on all parameter changes
            # TODO Improve
            self.reset()
        return changed_parameters

    def check_state(self, atoms):
        system_changes = super().check_state(atoms)
        if "pbc" in system_changes:
            # Ignore boundary conditions (we always use PBCs)
            system_changes.remove("pbc")
        return system_changes

    def calculate(self, atoms=None, properties=["energy"],
                  system_changes=all_changes):
        super().calculate(atoms=atoms, properties=properties,
                          system_changes=system_changes)

        # Write input file
        inputfile = self.label + ".json"
        self.write()

        # Run DFTK
        n_threads = self.parameters["n_threads"]
        n_mpi = self.parameters["n_mpi"]
        run_calculation(properties, inputfile, n_threads=n_threads,
                        n_mpi=n_mpi)

        # Read results
        self.read(self.label)


# ... and register it with ASE
register_calculator_class("DFTK", DFTK)



import asedftk

from numpy.testing import assert_allclose

from ase.build import bulk

ENERGY_PBE = -213.12688268374683  # eV
FORCES_PBE = [[0, 0, 0.], [0, 0, 0.]]


def test_silicon():
    silicon = bulk("Si")
    silicon.calc = asedftk.DFTK(xc="PBE", kpts=(3, 3, 3), ecut=190, scftol=1e-6,
                                mixing="SimpleMixing(α=1.3)")
    assert_allclose(silicon.get_potential_energy(), ENERGY_PBE, atol=1e-5)
    assert_allclose(silicon.get_forces(), FORCES_PBE, atol=1e-2)



import os

from asedftk import DFTK, julia

import ase
import ase.build


def dump_data():
    # Dump atoms data
    cell = [[3.21, 0.0, 0.0], [-1.605, 2.7799415461480477, 0.0], [0.0, 0.0, 5.21304]]
    positions = [[0.0, 0.0, 0.0], [0.0, 1.85329, 2.60652]]
    atoms = ase.Atoms(symbols='Mg2', pbc=True, cell=cell, positions=positions)
    calc = DFTK(label="mg")
    calc.atoms = atoms
    calc.write()

    calc = DFTK(label="si")
    calc.atoms = ase.build.bulk("Si")
    calc.write()


def test_juliatests():
    dump_data()

    # Run tests
    thisdir = os.path.dirname(__file__)
    julia(os.path.join(thisdir, "juliatests.jl"))
